import os

import pytest

from flaskel import middlewares, ObjectDict, TestClient
from flaskel.ext import (
    auth, builder, caching, cfremote, client_redis, cors, date_helper, errors,
    limit, logger, scheduler, sendmail, template, useragent
)
from flaskel.ext.crypto import argon2
from flaskel.ext.healthcheck import checks, health_checks
from flaskel.ext.sqlalchemy import db as sqlalchemy
from flaskel.extra.mobile_support import mobile_version, RedisStore
from flaskel.extra.stripe import payment_handler
from .blueprints import BLUEPRINTS, VIEWS

BASE_DIR = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
SKEL_DIR = os.path.join(BASE_DIR, 'skeleton')
CONF_DIR = os.path.join(SKEL_DIR, 'config')

CTS = ObjectDict(
    json='application/json',
    xml='application/xml',
    html='text/html',
    json_problem='application/problem+json',
    xml_problem='application/problem+xml',
    json_health='application/health+json',
)

HOSTS = ObjectDict(
    apitester="http://httpbin.org",
    fake="http://localhost"
)

SCHEMA_DIR = 'flaskel/scripts/skeleton/blueprints/api/static/schemas'

SCHEMAS = dict(
    JSONRPC=f"file://{SCHEMA_DIR}/jsonrpc.json",
    API_PROBLEM=f"file://{SCHEMA_DIR}/apiproblem.json",
    HEALTHCHECK=f"file://{SCHEMA_DIR}/healthcheck.json",
    OPENAPI3=f"file://{SCHEMA_DIR}/openapi3.json",
    ITEM={
        "$schema":    "http://json-schema.org/draft-07/schema#",
        "type":       "object",
        "required":   ["item"],
        "properties": {"item": {"type": "string"}}
    }
)

BASE_EXTENSIONS = {
    # "name":   (<extension>, parameters: dict)
    "cfremote":       (cfremote,),  # MUST be the first
    "logger":         (logger,),  # MUST be the second
    "template":       (template,),
    "builder":        (builder,),
    "date_helper":    (date_helper,),
    "cors":           (cors,),
    "database":       (sqlalchemy,),
    "limiter":        (limit.limiter,),
    "ip_ban":         (limit.ip_ban,),
    "cache":          (caching,),
    "errors":         (errors.error_handler, {
        "dispatcher": 'subdomain',
        "response":   builder.on_accept(strict=False),
        "normalizer": errors.ErrorNormalizer()
    }),
    "useragent":      (useragent,),
    "argon2":         (argon2,),
    "caching":        (caching,),
    "scheduler":      (scheduler,),
    "sendmail":       (sendmail.client_mail,),
    "redis":          (client_redis,),
    "health_checks":  (health_checks,),
    "jwt":            (auth.jwtm,),
    "ipban":          (limit.ip_ban,),
    "mobile_version": (mobile_version, dict(
        store=RedisStore(client_redis),
    )),
}

APISPEC = {
    "info":    {"version": None},
    "servers": [
        {
            "variables": {
                "context": {"default": None},
                "host":    {"default": None}
            }
        }
    ]
}

PROXIES = {
    "CONF": dict(
        host=HOSTS.apitester,
        url='/anything',
        headers={'k': 'v'},
        params={'k': 'v'}
    )
}


@pytest.fixture(scope='session')
def app_prod():
    # noinspection PyUnusedLocal
    def test_health_true(**kwargs):
        return True, None

    # noinspection PyUnusedLocal
    def test_health_false(**kwargs):
        return False, "error"

    health_checks.register('system')(checks.health_system)
    health_checks.register('mongo', db=sqlalchemy)(checks.health_mongo)
    health_checks.register('redis', db=sqlalchemy)(checks.health_redis)
    health_checks.register('sqlalchemy', db=sqlalchemy)(checks.health_sqlalchemy)

    return TestClient.get_app(
        conf=dict(
            DEBUG=True,
            BASIC_AUTH_USERNAME='username',
            BASIC_AUTH_PASSWORD='password',
            USER_AGENT_AUTO_PARSE=True,
            PREFERRED_URL_SCHEME='https',
            SCHEMAS=SCHEMAS,
            USE_X_SENDFILE=True,
            APIDOCS_ENABLED=True,
            APISPEC=APISPEC,
            PROXIES=PROXIES
        ),
        blueprints=BLUEPRINTS,
        extensions={
            **BASE_EXTENSIONS,
            **{
                "ipban":         None,
                "health_checks": (
                    health_checks, {'extensions': (
                        {'func': test_health_true, 'name': 'health_true'},
                        {'func': test_health_false},
                    )},
                ),
            }
        },
        middlewares=(
            middlewares.ForceHttps,
            middlewares.HTTPMethodOverride,
            middlewares.ReverseProxied,
            middlewares.RequestID,
        ),
        folders=["skeleton/blueprints/web/templates"],
        static_folder="skeleton/blueprints/web/static"
    )


@pytest.fixture(scope='session')
def app_dev():
    return TestClient.get_app(
        conf=dict(
            DEBUG=True,
            WSGI_WERKZEUG_LINT_ENABLED=True,
            WSGI_WERKZEUG_PROFILER_ENABLED=True,
            FLASK_ENV='development',
            JSONRPC_BATCH_MAX_REQUEST=2,
            SCHEMAS=SCHEMAS,
            PROXIES=PROXIES,
            STRIPE_SECRET_KEY='stripe_secret_key',
            STRIPE_PUBLIC_KEY='stripe_public_key'
        ),
        blueprints=(
            *BLUEPRINTS,
            *(None,),  # NB: needed to complete coverage
            *((None,),)  # NB: needed to complete coverage
        ),
        extensions={
            **BASE_EXTENSIONS,
            "empty":  (None, (None,)),  # NB: needed to complete coverage
            "ipban":  (limit.ip_ban, dict(nuisances=dict(string=["/phpmyadmin"]))),
            "stripe": (payment_handler,),
        },
        views=VIEWS,
        folders=["skeleton/blueprints/web/templates"],
        static_folder="skeleton/blueprints/web/static",
        after_request=(lambda x: x, lambda x: x),
        before_request=(lambda: None, lambda: None),
        version='1.0.0',
    )


@pytest.fixture(scope='session')
def testapp(app_prod):
    return app_prod.test_client()



# pragma: no cover
try:
    import redis
except ImportError:
    redis = None


class FlaskRedis:
    def __init__(self, app=None, client=None, **kwargs):
        """

        :param app:
        :param client:
        :param kwargs:
        """
        self._opts = kwargs
        self._client = client

        if app is not None:
            self.init_app(app)  # pragma: no cover

    @property
    def client(self):
        return self._client   # pragma: no cover

    def init_app(self, app):
        """

        :param app:
        """
        assert redis is not None, "redis must be installed"

        app.config.setdefault("REDIS_URL", "redis://localhost:6379/0")
        app.config.setdefault("REDIS_OPTS", self._opts)

        if not self._client:
            # noinspection PyUnresolvedReferences
            self._client = redis.StrictRedis.from_url(
                app.config["REDIS_URL"], **app.config["REDIS_OPTS"]
            )

        if not hasattr(app, "extensions"):
            app.extensions = {}  # pragma: no cover
        app.extensions['redis'] = self

    def __getattr__(self, name):
        return getattr(self._client, name)


client_redis = FlaskRedis()



import functools

import flask

from flaskel.ext.default import builder
from flaskel.flaskel import cap, httpcode
from flaskel.utils.batch import BatchExecutor, ThreadBatchExecutor


class HealthCheck:
    def __init__(self, app=None, **kwargs):
        """

        :param app:
        """
        self._executor = None
        self._health_checks = {}

        if app:
            self.init_app(app, **kwargs)  # pragma: no cover

    def init_app(self, app, bp=None, extensions=(), decorators=(), executor=None):
        """

        :param app: Flask app or blueprint instance
        :param bp: optional a blueprint instance on which register the healthcheck view
        :param extensions: functions and arguments that execute the checks
        :param decorators: functions that decorates the healthcheck view
        :param executor: batch executor class, must be subclass of BatchExecutor
        """
        self._executor = executor or ThreadBatchExecutor
        assert issubclass(self._executor, BatchExecutor)

        self.set_default_config(app)
        self.register_route(bp or app, decorators)
        self.register_extensions(app, extensions)

        if not hasattr(app, 'extensions'):
            app.extensions = dict()  # pragma: no cover
        app.extensions['healthcheck'] = self

    @staticmethod
    def set_default_config(app):
        app.config.setdefault('HEALTHCHECK_ABOUT_LINK', None)
        app.config.setdefault('HEALTHCHECK_VIEW_NAME', 'healthcheck')
        app.config.setdefault('HEALTHCHECK_PATH', '/healthcheck')
        app.config.setdefault('HEALTHCHECK_PARAM_KEY', 'checks')
        app.config.setdefault('HEALTHCHECK_PARAM_SEP', '+')
        app.config.setdefault('HEALTHCHECK_CONTENT_TYPE', 'application/health+json')

    def register_route(self, app, decorators=()):
        """

        :param app: Flask app or blueprint instance
        :param decorators: functions that decorates the healthcheck view
        """
        view = self.perform
        for decorator in decorators:
            view = decorator(view)

        conf = cap.config
        app.add_url_rule(conf['HEALTHCHECK_PATH'], conf['HEALTHCHECK_VIEW_NAME'], view_func=view)

    def register_extensions(self, app, extensions):
        """

        :param app: Flask app instance
        :param extensions: functions and arguments that execute the checks
        """
        for ex in extensions:
            try:
                func = ex.pop('func')
                self.register(**ex)(func)
            except Exception as exc:
                app.logger.exception(exc)
                app.logger.error("invalid healthcheck extension: %s", ex)

    @builder.response('json')
    def perform(self):
        healthy = True
        response = dict(
            status='pass', checks={},
            links={"about": cap.config['HEALTHCHECK_ABOUT_LINK']}
        )

        params = flask.request.args.get(cap.config['HEALTHCHECK_PARAM_KEY'])
        all_checks = set(self._health_checks.keys())
        if not params:
            params = all_checks
        else:
            params = params.split(cap.config['HEALTHCHECK_PARAM_SEP'])
            params = set(params).intersection(all_checks)

        # noinspection PyProtectedMember
        app = cap._get_current_object()
        params = list(params)
        tasks = [(self._health_checks.get(p), dict(app=app)) for p in params]
        resp = self._executor(tasks=tasks).run()

        for i, p in enumerate(params):
            state, message = resp[i]
            status = 'pass' if state else 'fail'
            response['checks'][p] = dict(status=status, output=message)
            if not state:
                healthy = False

        if not healthy:
            response['status'] = 'fail'

        status = httpcode.SUCCESS if healthy else httpcode.SERVICE_UNAVAILABLE
        return response, status, {'Content-Type': cap.config['HEALTHCHECK_CONTENT_TYPE']}

    def register(self, name=None, **kwargs):
        """

        :param name:
        :return:
        """

        def _register(func):
            nonlocal name
            name = name or func.__name__

            @functools.wraps(func)
            def wrapped():
                self._health_checks[name] = functools.partial(func, **kwargs)

            return wrapped()

        return _register


health_checks = HealthCheck()



try:
    from pyfcm import FCMNotification
    from pyfcm.errors import FCMError
    from sqlalchemy.exc import SQLAlchemyError
except ImportError:
    FCMNotification = FCMError = None
    SQLAlchemyError = None


class NotificationHandler:
    def __init__(self, model, session, provider=None, dry_run=False):
        """

        :param model: sqlalchemy model class, must have a token column
        :param session: sqlalchemy session
        :param provider: an object with app attribute reference of app instance (nor current_app)
                         only if async methods are used (for background tasks)
        :param dry_run: perform but not send notification (test only)
        """
        self.provider = provider
        self.session = session
        self.model = model
        self.dry_run = dry_run

        assert FCMNotification is not None, "you must install 'pyfcm'"
        assert SQLAlchemyError is not None, "you must install 'sqlalchemy'"

    def _with_app_context(self, f, *args, **kwargs):
        assert self.provider is not None, "you must pass a provider in order to user async methods"

        app = self.provider.app
        with app.app_context():
            return f(app, *args, **kwargs)

    def async_register_device(self, *args, **kwargs):
        return self._with_app_context(self.register_device, *args, **kwargs)

    def async_send_push_notification(self, *args, **kwargs):
        return self._with_app_context(self.send_push_notification, *args, **kwargs)

    def register_device(self, app, data):
        try:
            device = self.model.query.filter_by(token=data['token']).first()
            if device:
                self.session.merge(device.update(data))
            else:
                self.session.add(self.model(**data))
            self.session.commit()
        except SQLAlchemyError as exc:  # pragma: no cover
            app.logger.exception(exc)
            self.session.rollback()

    def send_push_notification(self, app, title, message, user_ids=None, tokens=None, **kwargs):
        """

        :param app: flask app instance
        :param title: notification title
        :param message: notification message
        :param user_ids: list of user ids (optional)
        :param tokens: list of fcm tokens (optional)
        :param kwargs: passed to notify_multiple_devices
        :return:
        """
        service = FCMNotification(api_key=app.config.FCM_API_KEY)

        if tokens is None:
            try:
                tokens = self.model.query \
                    .with_entities(self.model.token) \
                    .where(self.model.user_id.in_(user_ids)) \
                    .all()
            except SQLAlchemyError as exc:  # pragma: no cover
                app.logger.exception(exc)
                return

        tokens = [t[0] for t in tokens]
        # noinspection PyUnresolvedReferences
        rmax = FCMNotification.FCM_MAX_RECIPIENTS - 1
        if len(tokens) > rmax:
            tokens_set = [tokens[x:x + rmax] for x in range(0, len(tokens), rmax)]
        else:
            tokens_set = tokens

        kwargs.setdefault('sound', 'Default')
        kwargs.setdefault('dry_run', self.dry_run)

        for ts in tokens_set:
            try:
                res = service.notify_multiple_devices(
                    ts, message_title=title, message_body=message, **kwargs
                )
                app.logger.debug(res)
            except FCMError as exc:  # pragma: no cover
                app.logger.exception(exc)



try:  # pragma: no cover
    import holidays
except ImportError:  # pragma: no cover
    holidays = None

from datetime import datetime

from dateutil.parser import parse as date_parse

from flaskel.flaskel import cap


class Millis:
    seconds = 1000
    minute = seconds * 60
    hour = minute * 60
    day = hour * 24


class Seconds:
    millis = 1000
    minute = 60
    hour = minute * 60
    day = hour * 24


class Minutes:
    seconds = 60
    hour = 60
    day = hour * 24


class Day:
    hours = 24
    minutes = hours * 60
    seconds = minutes * 60


class DateHelper:
    def __init__(self, *args, **kwargs):
        self._holidays = holidays.CountryHoliday(*args, **kwargs)

    @property
    def holidays(self):
        return self._holidays  # pragma: no cover

    def all_holidays(self):
        return [d for d in self._holidays.items()]

    @staticmethod
    def country_holidays(*args, **kwargs):
        return holidays.CountryHoliday(*args, **kwargs)

    @staticmethod
    def is_weekend(curr_date, **kwargs):
        """

        :param curr_date: date instance or string date
        :param kwargs: passed to dateutil.parser.parse
        :return: boolean, True if is weekend
        """
        if isinstance(curr_date, str):
            curr_date = date_parse(curr_date, **kwargs)

        return curr_date.weekday() > 4

    @staticmethod
    def change_format(str_date, out_fmt, in_fmt=None, raise_exc=True):
        """

        :param str_date: input string date
        :param out_fmt: format output date
        :param in_fmt: format input date (optional: could be detected from string)
        :param raise_exc: raise or not exception (default True)
        :return: return formatted date
        """
        if not (raise_exc or str_date):
            return None  # pragma: no cover

        try:
            if in_fmt:
                date_time = datetime.strptime(str_date, in_fmt)
            else:
                date_time = date_parse(str_date)
            return date_time.strftime(out_fmt)
        except (ValueError, TypeError):
            if raise_exc is True:
                raise  # pragma: no cover


# for backwards compatibility
def from_iso_format(str_date, fmt, exc=True):
    return DateHelper.change_format(
        str_date, in_fmt=cap.config.DATE_ISO_FORMAT, out_fmt=fmt, raise_exc=exc
    )


# for backwards compatibility
def to_iso_format(str_date, fmt, exc=True):
    return DateHelper.change_format(
        str_date, in_fmt=fmt, out_fmt=cap.config.DATE_ISO_FORMAT, raise_exc=exc
    )



from .api import bp_api
from .auth import bp_auth
from .web import bp_web

BLUEPRINTS = (
    (bp_api,),
    (bp_web, {
        'url_prefix': '/'
    }),
    (bp_auth, {
        'url_prefix': '/auth'
    }),
)

VIEWS = (
    # (view, blueprint, dict options)
)



from flaskel import views
from flaskel.extra.mobile_support import MobileLoggerView, MobileReleaseView
from flaskel.views import apidoc, proxy, rpc
from .api import bp_api, resource, rpc as rpc_service
from .auth import bp_auth
from .test import bp_test
from .web import bp_web

jsonRPCView = rpc.JSONRPCView
rpc.JSONRPCView.load_from_object(rpc_service.MyJsonRPC())


class CustomProxy(proxy.TransparentProxyView):
    methods = ['POST']


BLUEPRINTS = (
    (bp_api,),
    (bp_test,),
    (bp_web, {
        'url_prefix': '/'
    }),
    (bp_auth, {
        'url_prefix': '/auth'
    }),
)

VIEWS = (
    (views.RenderTemplate, bp_web, dict(
        name='index',
        urls=['/'],
        template='index.html'
    )),
    (proxy.ConfProxyView, bp_api, dict(
        name='confproxy',
        urls=['/confproxy'],
        config_key='PROXIES.CONF'
    )),
    (CustomProxy, bp_api, dict(
        name='proxyview',
        urls=['/proxy'],
        method='POST',
        host='https://httpbin.org',
        url='/anything'
    )),
    (resource.APIResource, bp_api, dict(
        name='resource_api',
        url='/resources'
    )),
    (jsonRPCView, dict(
        name=jsonRPCView.default_view_name,
        url=jsonRPCView.default_url
    )),
    (apidoc.ApiDocTemplate, bp_api, dict(
        name=apidoc.ApiDocTemplate.default_view_name,
        urls=apidoc.ApiDocTemplate.default_urls
    )),
    (apidoc.ApiSpecTemplate, bp_api, dict(
        name=apidoc.ApiSpecTemplate.default_view_name,
        urls=apidoc.ApiSpecTemplate.default_urls
    )),
    (MobileReleaseView, bp_api, dict(
        name=MobileReleaseView.default_view_name,
        urls=MobileReleaseView.default_urls,
    )),
    (MobileLoggerView, bp_api, dict(
        name=MobileLoggerView.default_view_name,
        urls=MobileLoggerView.default_urls,
    )),
)



import os

import pytest

from flaskel import TestClient
from flaskel.utils import yaml
from scripts.cli import APP_CONFIG

yaml.setup_yaml_parser()

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
CONF_FILE = os.path.join(BASE_DIR, '..', 'config', 'conf.yaml')
CONFIG = yaml.load_yaml_file(CONF_FILE)

EXTRAS = dict(
    conf=CONFIG.app,
    template_folder="blueprints/web/templates",
    static_folder="blueprints/web/static"
)


@pytest.fixture(scope='session')
def testapp():
    return TestClient.get_app(**{**APP_CONFIG, **EXTRAS}).test_client()



import os
import time
from functools import partial

import flask
import werkzeug.exceptions

from flaskel import httpcode, misc, ObjectDict, uuid, yaml
from flaskel.http import batch, FlaskelHttp, FlaskelJsonRPC, HTTPClient, HTTPStatusError, rpc, useragent
from flaskel.tester import Asserter
from flaskel.utils import datetime, ExtProxy, schemas
from flaskel.utils.faker import DummyLogger
# noinspection PyUnresolvedReferences
from . import app_dev, app_prod, CTS, HOSTS, testapp

yaml.setup_yaml_parser()

url_for = partial(flask.url_for, _external=True)


def test_app_dev(app_dev):
    client = app_dev.test_client()
    res = client.get(url_for('web.index'))
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.html)
    Asserter.assert_equals(app_dev.config.FLASK_ENV, 'development')
    Asserter.assert_equals(app_dev.config.SECRET_KEY, 'fake_very_complex_string')
    res = client.get(url_for('test.test_https'))
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json.scheme, 'http')
    Asserter.assert_true(res.json.url_for.startswith('http'))


def test_api_resources(testapp):
    res = testapp.get(url_for('api.resource_api'), headers={'Accept': CTS.xml})
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.xml)

    res = testapp.get(url_for('api.resource_api', res_id=1))
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.json)

    res = testapp.get(url_for('api.resource_api', res_id=1, sub_resource='items'))
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.json)

    res = testapp.get(url_for('api.resource_api', res_id=1, sub_resource='not-found'))
    Asserter.assert_status_code(res, httpcode.NOT_FOUND)
    Asserter.assert_content_type(res, CTS.json_problem)
    Asserter.assert_schema(res.json, schemas.SCHEMAS.API_PROBLEM)
    Asserter.assert_equals(res.json.detail, 'not found')

    res = testapp.get(url_for('api.resource_api', res_id=1, sub_resource='not-found'))
    Asserter.assert_status_code(res, httpcode.TOO_MANY_REQUESTS)

    time.sleep(int(res.headers.get('Retry-After') or 0))  # check rate limit
    res = testapp.delete(url_for('api.resource_api', res_id=1))
    Asserter.assert_status_code(res, httpcode.NO_CONTENT)

    data = dict(item='test')
    res = testapp.put(url_for('api.resource_api', res_id=1), json=data)
    Asserter.assert_status_code(res)

    res = testapp.post(url_for('api.resource_api'), json=data)
    Asserter.assert_status_code(res, httpcode.CREATED)

    res = testapp.post(url_for('api.resource_api', res_id=1, sub_resource='items'), json=data)
    Asserter.assert_status_code(res, httpcode.CREATED)

    res = testapp.post(url_for('api.resource_api'), json={})
    Asserter.assert_allin(res.json.response.reason, ('cause', 'message', 'path'))
    Asserter.assert_status_code(res, httpcode.UNPROCESSABLE_ENTITY)


def test_api_cors(testapp):
    res = testapp.get(url_for('api.resource_api'))
    Asserter.assert_header(res, 'Access-Control-Allow-Origin', '*')


def test_dispatch_error_web(testapp):
    res = testapp.get('/web-page-not-found')
    Asserter.assert_status_code(res, httpcode.NOT_FOUND)
    Asserter.assert_content_type(res, CTS.html)


def test_dispatch_error_api(testapp):
    res = testapp.get(f"http://api.{testapp.application.config.SERVER_NAME}/not_found")
    Asserter.assert_status_code(res, httpcode.NOT_FOUND)
    Asserter.assert_content_type(res, CTS.json_problem)
    Asserter.assert_schema(res.json, schemas.SCHEMAS.API_PROBLEM)


def test_force_https(testapp):
    res = testapp.get(url_for('test.test_https'))
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json.scheme, 'https')
    Asserter.assert_true(res.json.url_for.startswith('https'))


def test_reverse_proxy(testapp):
    res = testapp.get(url_for('test.test_proxy'), headers={
        'X-Forwarded-Prefix': '/test'
    })
    Asserter.assert_status_code(res)
    Asserter.assert_true(bool(res.json.request_id))
    Asserter.assert_equals(res.json.script_name, '/test')
    Asserter.assert_equals(res.json.original.SCRIPT_NAME, '')


def test_secret_key_prod(testapp):
    Asserter.assert_equals(testapp.application.config.FLASK_ENV, 'production')
    Asserter.assert_true(os.path.isfile('.secret.key'))
    os.remove('.secret.key')


def test_method_override(testapp):
    res = testapp.post(
        url_for('test.method_override_post'),
        headers={'X-HTTP-Method-Override': 'PUT'}
    )
    Asserter.assert_status_code(res)

    res = testapp.post(f"{url_for('test.method_override_post')}?_method_override=PUT")
    Asserter.assert_status_code(res)


def test_converters(testapp):
    res = testapp.get(url_for('test.list_converter', data=['a', 'b', 'c']))
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json, 3)


def test_utils_get_json(testapp):
    res = testapp.post(url_for('test.get_invalid_json'))
    Asserter.assert_status_code(res, httpcode.BAD_REQUEST)


def test_utils_send_file(testapp):
    filename = 'MANIFEST.in'
    url = url_for('test.download') + "?filename={}"

    res = testapp.get(url.format(filename))
    Asserter.assert_status_code(res)
    Asserter.assert_header(res, 'Content-Disposition', f'attachment; filename={filename}')
    Asserter.assert_header(res, 'X-Accel-Redirect', f'./{filename}')
    Asserter.assert_true(res.headers.get('X-Sendfile').endswith(f'./{filename}'))
    Asserter.assert_equals(res.data, b'')

    res = testapp.get(url.format('nofile.txt'))
    Asserter.assert_status_code(res, httpcode.NOT_FOUND)
    Asserter.assert_content_type(res, CTS.html)


def test_utils_uuid(testapp):
    res = testapp.get(url_for('test.return_uuid'))
    Asserter.assert_false(uuid.check_uuid('fake uuid'))
    Asserter.assert_true(uuid.check_uuid(uuid.get_uuid(hexify=False)))
    Asserter.assert_true(uuid.check_uuid(res.json.get('uuid1'), ver=1))
    Asserter.assert_true(uuid.check_uuid(res.json.get('uuid3'), ver=3))
    Asserter.assert_true(uuid.check_uuid(res.json.get('uuid4')))
    Asserter.assert_true(uuid.check_uuid(res.json.get('uuid5'), ver=5))


def test_crypto(testapp):
    passwd = 'my-favourite-password'
    crypto = ExtProxy('argon2')
    res = testapp.get(url_for('test.crypt', passwd=passwd))
    Asserter.assert_true(crypto.verify_hash(res.data, passwd))
    Asserter.assert_false(crypto.verify_hash(res.data, "wrong-pass"))


def test_utils_http_client_simple(testapp):
    with testapp.application.test_request_context():
        api = FlaskelHttp(HOSTS.apitester, token='pippo', dump_body=True)
        res = api.delete('/status/202')
        Asserter.assert_equals(res.status, httpcode.ACCEPTED)
        res = api.patch('/status/400')
        Asserter.assert_equals(res.status, httpcode.BAD_REQUEST)


def test_utils_http_client_exception(testapp):
    logger = testapp.application.logger
    api = HTTPClient(HOSTS.apitester, token='pippo', raise_on_exc=True, logger=logger)
    fake_api = HTTPClient(HOSTS.fake, username='test', password='test', logger=logger)

    res = fake_api.put('/', timeout=0.1)
    Asserter.assert_equals(res.status, httpcode.SERVICE_UNAVAILABLE)
    try:
        api.request('/status/500', 'PUT')
    except HTTPStatusError as exc:
        Asserter.assert_equals(exc.response.status_code, httpcode.INTERNAL_SERVER_ERROR)
    try:
        fake_api.request('/', timeout=0.1)
    except werkzeug.exceptions.HTTPException as exc:
        Asserter.assert_equals(exc.code, httpcode.INTERNAL_SERVER_ERROR)


def test_utils_http_client_filename():
    filename = "pippo.txt"
    hdr = 'Content-Disposition'
    param = {hdr: None}

    api = HTTPClient(HOSTS.apitester, dump_body=True)
    res = api.get('/not-found')
    Asserter.assert_status_code(res, httpcode.NOT_FOUND)

    param[hdr] = f"attachment; filename={filename}"
    res = api.get('/response-headers', params=param)
    Asserter.assert_equals(api.response_filename(res.headers), filename)

    param[hdr] = f"filename={filename}"
    res = api.post('/response-headers', params=param)
    Asserter.assert_equals(api.response_filename(res.headers), filename)

    param[hdr] = filename
    res = api.post('/response-headers', params=param)
    Asserter.assert_none(api.response_filename(res.headers))


def test_http_client_batch(testapp):
    with testapp.application.test_request_context():
        responses = batch.FlaskelHTTPBatch(logger=DummyLogger(), dump_body=(True, False)).request([
            dict(url=f"{HOSTS.apitester}/anything", method="GET", headers={"HDR1": "HDR1"}),
            dict(url=f"{HOSTS.apitester}/status/{httpcode.NOT_FOUND}", method="GET"),
            dict(url=HOSTS.fake, method='GET', timeout=0.1),
        ])
    Asserter.assert_equals(responses[0].body.headers.Hdr1, 'HDR1')
    Asserter.assert_equals(responses[1].status, httpcode.NOT_FOUND)
    Asserter.assert_equals(responses[2].status, httpcode.SERVICE_UNAVAILABLE)


def test_utils_http_jsonrpc_client(testapp):
    params = dict(a=1, b=2)
    with testapp.application.test_request_context():
        api = FlaskelJsonRPC(HOSTS.apitester, "/anything")
        res = api.request('method.test', params=params)
        Asserter.assert_equals(res.json.jsonrpc, '2.0')
        Asserter.assert_equals(res.json.id, api.request_id)
        Asserter.assert_equals(res.json.params, params)


def test_healthcheck(testapp):
    res = testapp.get(testapp.application.config.HEALTHCHECK_PATH)
    Asserter.assert_status_code(res, httpcode.SERVICE_UNAVAILABLE)
    Asserter.assert_content_type(res, CTS.json_health)
    Asserter.assert_schema(res.json, schemas.SCHEMAS.HEALTHCHECK)
    Asserter.assert_allin(res.json.checks.keys(), ('mongo', 'redis', 'sqlalchemy', 'system'))


def test_api_jsonrpc_success(app_dev):
    call_id = 1
    url = url_for('jsonrpc')
    testapp = app_dev.test_client()
    res = testapp.jsonrpc(url, method="MyJsonRPC.testAction1", call_id=call_id)
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, schemas.JSONSchema.load_from_file(schemas.SCHEMAS.JSONRPC)['RESPONSE'])
    Asserter.assert_equals(res.json.id, call_id)
    Asserter.assert_true(res.json.result.action1)

    res = testapp.jsonrpc(url, method="MyJsonRPC.testAction2")
    Asserter.assert_status_code(res, httpcode.NO_CONTENT)


def test_api_jsonrpc_error(app_dev):
    call_id = 1
    url = url_for('jsonrpc')
    testapp = app_dev.test_client()
    response_schema = schemas.JSONSchema.load_from_file(schemas.SCHEMAS.JSONRPC)['RESPONSE']
    headers = dict(headers={app_dev.config.LIMITER.BYPASS_KEY: app_dev.config.LIMITER.BYPASS_VALUE})

    res = testapp.jsonrpc(url, method="NotFoundMethod", call_id=call_id, **headers)
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json.error.code, rpc.RPCMethodNotFound().code)
    Asserter.assert_schema(res.json, response_schema)
    Asserter.assert_equals(res.json.id, call_id)
    Asserter.assert_true(res.json.error.message)

    res = testapp.jsonrpc(url, json={}, **headers)
    Asserter.assert_status_code(res, httpcode.BAD_REQUEST)
    Asserter.assert_schema(res.json, response_schema)
    Asserter.assert_equals(res.json.error.code, rpc.RPCParseError().code)

    res = testapp.jsonrpc(url, json={"params": None}, **headers)
    Asserter.assert_status_code(res, httpcode.BAD_REQUEST)
    Asserter.assert_schema(res.json, response_schema)
    Asserter.assert_equals(res.json.error.code, rpc.RPCInvalidRequest().code)

    res = testapp.jsonrpc(url, json={"jsonrpc": 1, 'method': "MyJsonRPC.testAction1"}, **headers)
    Asserter.assert_status_code(res, httpcode.BAD_REQUEST)
    Asserter.assert_schema(res.json, response_schema)
    Asserter.assert_equals(res.json.error.code, rpc.RPCInvalidRequest().code)

    res = testapp.jsonrpc(url, method="MyJsonRPC.testInternalError", call_id=call_id, **headers)
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, response_schema)
    Asserter.assert_equals(res.json.error.code, rpc.RPCInternalError().code)


def test_api_jsonrpc_params(app_dev):
    url = url_for('jsonrpc')
    method = "MyJsonRPC.testInvalidParams"
    testapp = app_dev.test_client()
    response_schema = schemas.JSONSchema.load_from_file(schemas.SCHEMAS.JSONRPC)['RESPONSE']
    headers = dict(headers={app_dev.config.LIMITER.BYPASS_KEY: app_dev.config.LIMITER.BYPASS_VALUE})

    res = testapp.jsonrpc(url, method=method, call_id=1, params={"param": "testparam"}, **headers)
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, response_schema)
    Asserter.assert_not_in('error', res.json)

    res = testapp.jsonrpc(url, method=method, call_id=1, params={"params": None}, **headers)
    Asserter.assert_status_code(res)
    Asserter.assert_schema(res.json, response_schema)
    Asserter.assert_equals(res.json.error.code, rpc.RPCInvalidParams().code)


def test_api_jsonrpc_batch(app_dev):
    url = url_for('jsonrpc')
    testapp = app_dev.test_client()
    headers = dict(headers={app_dev.config.LIMITER.BYPASS_KEY: app_dev.config.LIMITER.BYPASS_VALUE})
    res = testapp.jsonrpc_batch(url, requests=(
        dict(method="MyJsonRPC.testAction1", call_id=1, params={}),
        dict(method="MyJsonRPC.NotFoundMethod", call_id=2),
    ), **headers)
    Asserter.assert_status_code(res, httpcode.MULTI_STATUS)
    Asserter.assert_true(res.json[0].result.action1)
    Asserter.assert_equals(res.json[1].error.code, rpc.RPCMethodNotFound().code)

    res = testapp.jsonrpc_batch(url, requests=(
        dict(method="MyJsonRPC.testAction1", call_id=1, params={}),
        dict(method="MyJsonRPC.NotFoundMethod", call_id=2),
        dict(method="MyJsonRPC.NotFoundMethod", call_id=3),
    ), **headers)
    Asserter.assert_status_code(res, httpcode.REQUEST_ENTITY_TOO_LARGE)


def test_api_jsonrpc_notification(app_dev):
    url = url_for('jsonrpc')
    testapp = app_dev.test_client()
    headers = dict(headers={app_dev.config.LIMITER.BYPASS_KEY: app_dev.config.LIMITER.BYPASS_VALUE})
    res = testapp.jsonrpc_batch(url, requests=(
        dict(method="MyJsonRPC.testAction1", params={}),
        dict(method="MyJsonRPC.NotFoundMethod"),
    ), **headers)
    Asserter.assert_status_code(res, httpcode.NO_CONTENT)

    res = testapp.jsonrpc_batch(url, requests=(
        dict(method="MyJsonRPC.testAction1", call_id=1, params={}),
        dict(method="MyJsonRPC.NotFoundMethod"),
    ), **headers)
    Asserter.assert_status_code(res)
    Asserter.assert_equals(len(res.json), 1)


def test_useragent(testapp):
    res = testapp.get(url_for('test.useragent'))
    Asserter.assert_status_code(res)
    Asserter.assert_allin(res.json.keys(), ('browser', 'device', 'os', 'raw'))
    ua_string = """
        Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) 
        Chrome/70.0.3538.77 Safari/537.36
    """
    res = useragent.UserAgent().parse(ua_string)
    Asserter.assert_equals(res.raw, ua_string)


def test_utils_date_conversion():
    exc = False
    fmt = "%d %B %Y %I:%M %p"
    iso_date = '2020-12-28T19:53:00'
    not_iso_date = '28 December 2020 07:53 PM'
    invalid_date = 'invalid_date'

    res = datetime.from_iso_format(iso_date, fmt, exc=exc)
    Asserter.assert_true(res)
    Asserter.assert_equals(not_iso_date, res)

    res = datetime.to_iso_format(not_iso_date, fmt, exc=exc)
    Asserter.assert_true(res)
    Asserter.assert_equals(iso_date, res)

    Asserter.assert_none(datetime.from_iso_format(invalid_date, fmt, exc=exc))
    Asserter.assert_none(datetime.to_iso_format(invalid_date, fmt, exc=exc))


def test_correlation_id(testapp):
    cap = testapp.application
    request_id_header = cap.config.REQUEST_ID_HEADER

    res = testapp.get('/')
    Asserter.assert_header(res, request_id_header, is_in=True)
    Asserter.assert_type(res.headers[request_id_header], str)

    req_id = uuid.get_uuid()
    res = testapp.get('/', headers={request_id_header: req_id})
    Asserter.assert_header(res, request_id_header, req_id)

    req_id = 'invalid_req_id'
    res = testapp.get('/', headers={request_id_header: req_id})
    Asserter.assert_different(res.headers[request_id_header], req_id)


def test_jwt(testapp):
    config = testapp.application.config
    bypass = {config.LIMITER.BYPASS_KEY: config.LIMITER.BYPASS_VALUE}
    tokens = testapp.post(
        url_for('auth.access_token'),
        json=dict(email='email', password='password')
    )
    Asserter.assert_status_code(tokens)
    Asserter.assert_allin(tokens.json.keys(), (
        'access_token', 'refresh_token', 'expires_in', 'issued_at', 'token_type', 'scope'
    ))

    res = testapp.post(
        url_for('auth.refresh_token'),
        headers={'Authorization': f"Bearer {tokens.json.refresh_token}", **bypass}
    )
    Asserter.assert_status_code(res)
    Asserter.assert_allin(res.json.keys(), (
        'access_token', 'expires_in', 'issued_at', 'token_type', 'scope'
    ))

    check = testapp.get(
        url_for('auth.check_token'),
        headers={'Authorization': f"Bearer {res.json.access_token}", **bypass}
    )
    Asserter.assert_status_code(check)

    unauth = testapp.get(
        url_for('auth.check_token'),
        headers={'Authorization': f"Bearer invalid-token", **bypass}
    )
    Asserter.assert_status_code(unauth, httpcode.UNAUTHORIZED)

    revoked = testapp.post(
        url_for('auth.revoke_token'),
        headers={'Authorization': f"Bearer {tokens.json.access_token}", **bypass},
        json={'access_token': tokens.json.access_token, 'refresh_token': tokens.json.refresh_token}
    )
    Asserter.assert_status_code(revoked, httpcode.NO_CONTENT)

    res = testapp.get(
        url_for('auth.check_token'),
        headers={'Authorization': f"Bearer {tokens.json.access_token}", **bypass}
    )
    Asserter.assert_status_code(res, httpcode.UNAUTHORIZED)


def test_http_status():
    Asserter.assert_true(httpcode.is_informational(httpcode.PROCESSING))
    Asserter.assert_true(httpcode.is_success(httpcode.CREATED))
    Asserter.assert_false(httpcode.is_success(httpcode.MULTIPLE_CHOICES))
    Asserter.assert_true(httpcode.is_redirection(httpcode.SEE_OTHER))
    Asserter.assert_false(httpcode.is_redirection(httpcode.BAD_REQUEST))
    Asserter.assert_true(httpcode.is_client_error(httpcode.UNAUTHORIZED))
    Asserter.assert_false(httpcode.is_client_error(httpcode.INTERNAL_SERVER_ERROR))
    Asserter.assert_true(httpcode.is_server_error(httpcode.NOT_IMPLEMENTED))
    Asserter.assert_false(httpcode.is_server_error(httpcode.NOT_MODIFIED))
    Asserter.assert_true(httpcode.is_ok(httpcode.FOUND))
    Asserter.assert_false(httpcode.is_ko(httpcode.SUCCESS))
    Asserter.assert_status_code(ObjectDict(status=201), code=(201, 202, 203), is_in=True)
    Asserter.assert_status_code(ObjectDict(status=201), code=(200, 299), in_range=True)
    Asserter.assert_status_code(ObjectDict(status=400), code=300, greater=True)
    Asserter.assert_status_code(ObjectDict(status=200), code=300, less=True)


def test_proxyview(testapp):
    res = testapp.post(f"{url_for('api.proxyview')}?p=v1", json={'test': 'test'})
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json.json.test, 'test')
    Asserter.assert_equals(res.json.args.p, 'v1')

    res = testapp.get(f"{url_for('api.confproxy')}")
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json.headers.K, 'v')
    Asserter.assert_equals(res.json.args.k, 'v')


def test_apidoc(testapp):
    res = testapp.get(url_for('api.apidocs'))
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.html)

    res = testapp.get(url_for('api.apispec'))
    Asserter.assert_status_code(res)
    Asserter.assert_content_type(res, CTS.json)
    Asserter.assert_equals(res.json.info.version, '1.0.0')
    Asserter.assert_equals(res.json.servers[0].variables.context.default, '/')
    Asserter.assert_equals(res.json.servers[0].variables.host.default, 'https://127.0.0.1:5000')


def test_mobile_version(app_dev):
    testapp = app_dev.test_client()
    headers = {app_dev.config.LIMITER.BYPASS_KEY: app_dev.config.LIMITER.BYPASS_VALUE}
    res = testapp.get(url_for('web.index'), headers=headers)
    Asserter.assert_header(res, 'X-Upgrade-Needed', '0')
    Asserter.assert_header(res, 'X-Api-Version', '1.0.0')

    res = testapp.get('/web-page-not-found')
    Asserter.assert_not_in('X-Upgrade-Needed', list(res.headers.keys()))
    Asserter.assert_not_in('X-Api-Version', list(res.headers.keys()))


def test_mobile_views(app_dev):
    testapp = app_dev.test_client()
    url = url_for('api.mobile_logger')
    data = {'stacktrace': "exception"}

    res = testapp.post(url, json=data)
    print(res.json)
    Asserter.assert_status_code(res, httpcode.NO_CONTENT)

    res = testapp.post(f"{url}?debug", json=data)
    Asserter.assert_status_code(res)
    Asserter.assert_equals(res.json, data)


def test_ipban(app_dev):  # must be last test on dev app
    res = None
    conf = app_dev.config
    ipban = ExtProxy('ipban')
    ipban.remove_whitelist('127.0.0.1')
    for i in range(0, conf.IPBAN_COUNT + 2):
        testapp = app_dev.test_client()
        res = testapp.get(f"{conf.SERVER_NAME}/phpmyadmin")

    Asserter.assert_status_code(res, httpcode.FORBIDDEN)


def test_misc():
    Asserter.assert_less(1, 2)
    Asserter.assert_greater(2, 1)
    Asserter.assert_range(2, (1, 3))
    Asserter.assert_occurrence('abacca', r'a', 3)
    Asserter.assert_occurrence('abacca', r'a', 4, less=True)
    Asserter.assert_occurrence('abacca', r'a', 2, greater=True)
    Asserter.assert_true(misc.to_float("1.1"))
    Asserter.assert_none(misc.to_float("1,1"))
    Asserter.assert_true(misc.to_int("1"))
    Asserter.assert_none(misc.to_int("1,1"))
    Asserter.assert_true(misc.parse_value('true'))
    Asserter.assert_equals("test", misc.parse_value('test'))
    Asserter.assert_equals(misc.to_int, misc.import_from_module("flaskel.utils.misc:to_int"))



from flask import current_app as cap, json


class BaseHTTPDumper:
    @classmethod
    def dump_headers(cls, hdr, only=()):
        """
        dumps http headers: useful for logging

        :param hdr: headers' dictionary
        :param only: list of headers key to dump
        :return: string representation of headers
                k1: v1
                k2: v2
        """
        hdr = hdr or {}
        if only:
            hdr = {k: hdr[k] for k in only if k in hdr}

        headers = '\n\t'.join(f"{k}: {v}" for k, v in hdr.items())
        return headers.strip()

    @classmethod
    def response_filename(cls, headers):
        """
        get filename from Content-Disposition header.
        i.e.: attachment; filename="<file name.ext>"

        :param headers: http headers dict
        :return: only the file name
        """
        headers = headers or {}
        hdr = headers.get('Content-Disposition')
        if not hdr:
            return None  # pragma: no cover

        tmp = hdr.split(';')
        hdr = tmp[1] if len(tmp) > 1 else tmp[0]
        tmp = hdr.split('=')

        if len(tmp) > 1:
            return tmp[1].strip('"').lstrip('<').rstrip('>')
        else:
            return None

    @classmethod
    def dump_request(cls, req, dump_body=None, only_hdr=()):
        """
        dump http request: useful for logging

        :param req: Request instance
        :param dump_body: flag to enable or disable dump of request's body (overrides default)
        :param only_hdr: dump only a subset of headers
        :return: prettified representation of input as string
        """
        body = ''
        if dump_body is True:
            body = getattr(req, 'json', None)
            if body:
                body = json.dumps(body)
            else:
                body = getattr(req, 'body', getattr(req, 'data', None))
                if isinstance(body, (bytes, bytearray)):
                    body = 'Binary body not dumped'
            body = f"\nbody:\n{body}" if body else ''

        headers = cls.dump_headers(req.headers, only=only_hdr)
        headers = f"\nheaders:\n\t{headers}" if headers else ''
        return f"requested {req.method} {req.url}{headers}{body}"

    @classmethod
    def dump_response(cls, resp, dump_body=None, only_hdr=()):
        """
        dump http response: useful for logging

        :param resp: Response instance
        :param dump_body: flag to enable or disable dump of response's body (overrides default)
        :param only_hdr: dump only a subset of headers
        :return: prettified representation of input as string
        """
        body = ''
        if dump_body is True:
            body = cls.response_filename(resp.headers) or resp.text
            body = f"\nbody:\n{body}" if body else ''

        try:
            seconds = resp.elapsed.total_seconds()
        except AttributeError:  # because aiohttp.ClientResponse has not elapsed attribute
            seconds = 'N/A'

        headers = cls.dump_headers(resp.headers, only=only_hdr)
        headers = f"\nheaders:\n\t{headers}" if headers else ''
        try:
            status = resp.status_code
        except AttributeError:
            status = getattr(resp, 'status', None)

        return f"response time: {seconds} - status code: {status}{headers}{body}"


class LazyHTTPDumper(BaseHTTPDumper):
    @classmethod
    def dump_request(cls, req, *args, **kwargs):
        dump = super().dump_request

        class DumpRequest:
            def __str__(self):
                return dump(req, *args, **kwargs)

        return DumpRequest()

    @classmethod
    def dump_response(cls, resp, *args, **kwargs):
        dump = super().dump_response

        class DumpResponse:
            def __str__(self):
                return dump(resp, *args, **kwargs)

        return DumpResponse()


class FlaskelHTTPDumper(LazyHTTPDumper):
    @classmethod
    def dump_request(cls, req, dump_body=None, **kwargs):
        h = cap.config.LOG_REQ_HEADERS
        return super().dump_request(req, dump_body, only_hdr=h, **kwargs)

    @classmethod
    def dump_response(cls, resp, dump_body=None, **kwargs):
        h = cap.config.LOG_RESP_HEADERS
        return super().dump_response(resp, dump_body, only_hdr=h, **kwargs)



"""
Flaskel
-------
"""
import os
import re
import sys

from setuptools import find_packages as base_find_packages, setup
from setuptools.command.test import test

LICENSE = 'MIT'
PLATFORMS = 'any'
PYTHON_VERSION = '>3.6'
URL = 'https://github.com/cs91chris/flaskel'
DESCRIPTION = 'Skeleton for flask applications'

PKG_NAME = 'flaskel'
PKG_TEST = 'tests'
PKG_SCRIPTS = f'{PKG_NAME}.scripts'

ENTRY_POINTS = dict(
    console_scripts=[
        f"{PKG_NAME}={PKG_SCRIPTS}.cli:cli",
    ],
)

REQUIRES = [
    "Flask==1.1.*",
    "Flask-APScheduler==1.12.*",
    "Flask-Caching==1.10.*",
    "Flask-CloudflareRemote==1.1.*",
    "Flask-Cors==3.0.*",
    "Flask-ErrorsHandler==4.0.*",
    "Flask-HTTPAuth==4.2.*",
    "Flask-JWT-Extended==4.1.*",
    "Flask-Limiter==1.4.*",
    "Flask-Logify==2.3.*",
    "Flask-Mail==0.9.*",
    "Flask-ResponseBuilder==2.0.*",
    "Flask-SQLAlchemy==2.5.*",
    "Flask-TemplateSupport==2.0.*",
    "aiohttp==3.7.*",
    "argon2-cffi==20.1.*",
    "coverage==5.5.*",
    "nest-asyncio==1.5.*",
    "psutil==5.8.*",
    "pytest==6.2.*",
    "pytest-cov==2.11.*",
    "python-dateutil==2.8.*",
    "python-decouple==3.4.*",
    "PyYAML==5.4.*",
    "requests==2.25.*",
    "jsonschema==3.2.*",
    "simplejson==3.*",
    "sqlalchemy-schemadisplay==1.3.*",
    "user-agents==2.2.*",
    "webargs==8.0.*",
    "redis==3.5.*",
    "hiredis==2.0.*",
]

BASE_PATH = os.path.dirname(__file__)
SKEL_DIR = os.path.join(*PKG_SCRIPTS.split('.'), 'skeleton')
VERSION_FILE = os.path.join(BASE_PATH, PKG_NAME, 'version.py')

try:
    # must be after setuptools
    # noinspection PyPackageRequirements
    from Cython.Build import cythonize as base_cythonize
    # noinspection PyPackageRequirements
    import Cython.Compiler.Options as cython_options

    cython_options.docstrings = False
except ImportError:
    cython_options = None
    base_cythonize = None

if "--cythonize" not in sys.argv:
    base_cythonize = None
else:
    sys.argv.remove("--cythonize")


def ext_paths(root_dir, exclude=()):
    paths = []
    for root, dirs, files in os.walk(root_dir):
        for filename in files:
            file_path = os.path.join(root, filename)
            file_ext = filename.split('.')[-1]
            if file_path in exclude or file_ext not in ('py', 'pyx'):
                continue

            paths.append(file_path)
    return paths


def read(file):
    with open(os.path.join(BASE_PATH, file)) as f:
        return f.read()


def grep(file, name):
    value, = re.findall(fr"{name}\W*=\W*'([^']+)'", read(file))
    return value


def readme(file):
    try:
        return read(file)
    except OSError as exc:
        print(str(exc), file=sys.stderr)


def skeleton_files():
    paths = []
    for path, _, filenames in os.walk(SKEL_DIR):
        for f in filenames:
            paths.append(os.path.join(path, f))
    return paths


class PyTest(test):
    def finalize_options(self):
        test.finalize_options(self)

    def run_tests(self):
        import pytest
        sys.exit(pytest.main([PKG_TEST]))


def cythonize(paths):
    if base_cythonize is not None:
        return base_cythonize(paths, language_level=3)


def find_packages():
    if base_cythonize is not None:
        return [PKG_SCRIPTS]
    return base_find_packages()


setup(
    name='Flaskel',
    url=URL,
    license=LICENSE,
    description=DESCRIPTION,
    platforms=PLATFORMS,
    python_requires=PYTHON_VERSION,
    long_description=readme('README.rst'),
    version=grep(VERSION_FILE, '__version__'),
    author=grep(VERSION_FILE, '__author_name__'),
    author_email=grep(VERSION_FILE, '__author_email__'),
    zip_safe=False,
    include_package_data=True,
    packages=find_packages(),
    ext_modules=cythonize(ext_paths(PKG_NAME, skeleton_files())),
    test_suite=PKG_TEST,
    entry_points=ENTRY_POINTS,
    cmdclass=dict(test=PyTest),
    install_requires=REQUIRES,
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    ]
)



from flask_sqlalchemy import event
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_property

from flaskel import ExtProxy, ObjectDict
from flaskel.ext.sqlalchemy import db


class StandardMixin:
    def __init__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """

    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    created_at = db.Column(db.DateTime, server_default=db.func.now())
    updated_at = db.Column(db.DateTime, onupdate=db.func.now())


class CatalogMixin:
    def __init__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """

    @declared_attr
    def __table_args__(self):
        return db.UniqueConstraint('label', 'type_id'),

    id = db.Column(db.Integer, primary_key=True)
    label = db.Column(db.String(100))
    type_id = db.Column(db.Integer)
    description = db.Column(db.String(250))

    def __str__(self):
        return self.label

    def to_dict(self, **kwargs):
        return ObjectDict(
            id=self.id,
            label=self.label,
            description=self.description,
            typeId=self.type_id,
            **kwargs
        )


class CatalogXMixin(CatalogMixin):
    @declared_attr
    def __table_args__(self):
        return db.UniqueConstraint('code', 'type_id'),

    code = db.Column(db.String(20))
    order_id = db.Column(db.Integer, index=True)

    order_by = order_id.asc()

    def __str__(self):
        return f"<{self.code} - {self.label}>"

    def to_dict(self, **kwargs):
        return super().to_dict(
            code=self.code,
            orderId=self.order_id,
            **kwargs
        )


class LoaderMixin:
    values = ()

    def __init__(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        """

    @classmethod
    def load_values(cls, *args, **kwargs):
        for d in cls.values:
            db.session.add(cls(**d))
        db.session.commit()

    @classmethod
    def register_loader(cls):
        # noinspection PyUnresolvedReferences
        event.listen(cls.__table__, 'after_create', cls.load_values)


class UserMixin(StandardMixin):
    _hasher = ExtProxy('argon2')
    _password = db.Column('password', db.String(128), nullable=False)

    email = db.Column(db.String(255), unique=True, nullable=False)

    @hybrid_property
    def password(self):
        return self._password

    @password.setter
    def password(self, password):
        self._password = self._hasher.generate_hash(password)

    def check_password(self, password):
        return self._hasher.verify_hash(self._password, password)

    def to_dict(self, restricted=False):
        return ObjectDict(
            id=self.id,
            email=self.email
        )



from flaskel.ext import (
    caching, client_mail, client_redis, date_helper,
    default, errors, limit, scheduler, useragent
)
from flaskel.ext.auth import jwtm
from flaskel.ext.crypto import argon2
from flaskel.ext.healthcheck import health_checks
from flaskel.ext.sqlalchemy import db as sqlalchemy
from flaskel.extra.mobile_support import mobile_version, RedisStore
from flaskel.extra.stripe import payment_handler


class OPTS:
    cors = dict(
        resources={r'/*': {'origins': '*'}}
    )
    mobile_version = dict(
        store=RedisStore(client_redis)
    )
    errors = dict(
        dispatcher='subdomain',
        response=default.builder.on_accept(strict=False),
    )


EXTENSIONS = {
    # "name":   (<extension>, parameters: dict)
    "cfremote":       (default.cfremote,),  # MUST be the first
    "logger":         (default.logger,),  # MUST be the second
    "template":       (default.template,),
    "builder":        (default.builder,),
    "date_helper":    (date_helper,),
    "cors":           (default.cors, OPTS.cors),
    "database":       (sqlalchemy,),
    "limiter":        (limit.limiter,),
    "ip_ban":         (limit.ip_ban,),
    "cache":          (caching,),
    "errors":         (errors.error_handler, OPTS.errors),
    "useragent":      (useragent,),
    "argon2":         (argon2,),
    "caching":        (caching,),
    "scheduler":      (scheduler,),
    "sendmail":       (client_mail,),
    "redis":          (client_redis,),
    "health_checks":  (health_checks,),
    "jwt":            (jwtm,),
    "ipban":          (limit.ip_ban,),
    "stripe":         (payment_handler,),
    "mobile_version": (mobile_version, OPTS.mobile_version),
}



import atexit
import logging

from apscheduler import events
from apscheduler.schedulers.blocking import BlockingScheduler
from flask_apscheduler import APScheduler

from flaskel.utils.uuid import get_uuid

try:
    import fcntl
except ImportError:
    fcntl = None


class APJobs(APScheduler):
    def init_app(self, app, test_sync=True):
        app.config.setdefault('SCHEDULER_AUTO_START', False)
        app.config.setdefault('SCHEDULER_PATCH_MULTIPROCESS', True)

        if app.config.SCHEDULER_PATCH_MULTIPROCESS is True:
            if fcntl is None:
                app.logger.warning('fcntl not supported on this platform')
            elif not self._set_lock():
                return  # pragma: no cover

        super().init_app(app)

        # if app is in testing mode execute tasks synchronously
        if test_sync and app.testing:
            self._scheduler = BlockingScheduler()

        if app.debug:
            logger = logging.getLogger('apscheduler')
            logger.setLevel(logging.DEBUG)
            logger = logging.getLogger('flask_apscheduler')
            logger.setLevel(logging.DEBUG)

        if not hasattr(app, 'extensions'):
            app.extensions = dict()  # pragma: no cover
        app.extensions['scheduler'] = self

        if app.config.SCHEDULER_AUTO_START is True:
            self.start()

    def add_job(self, func, kwargs=None, **kw):
        """

        :param func:
        :param kwargs:
        :return:
        """
        kwargs = kwargs or {}
        kw['kwargs'] = kwargs
        job_id = get_uuid()
        job = super().add_job(id=job_id, func=func, **kw)
        self.app.logger.debug('added job %s: %s', func, job_id)
        return job

    @staticmethod
    def _set_lock():  # pragma: no cover
        """
        ensure that only one worker starts the scheduler
        :return: (boolean) True -> scheduler can start
        """
        f = None

        def unlock():
            try:
                fcntl.flock(f, fcntl.LOCK_UN)
                f.close()
            except OSError:
                pass

        try:
            f = open(".scheduler.lock", "wb")
            fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)
            atexit.register(unlock)
            return True
        except OSError:
            if f is not None:
                try:
                    f.close()
                except OSError:
                    pass
        return False


scheduler = APJobs()


def exception_listener(event):
    logger = scheduler.app.logger
    if event.exception:
        logger.error(f"An error occurred when executing job: {event.job_id}")
        logger.exception(event.exception)
        logger.error(event.traceback)
    else:
        logger.debug(f"successfully executed job: {event.job_id}")


scheduler.add_listener(
    exception_listener,
    events.EVENT_JOB_EXECUTED | events.EVENT_JOB_ERROR
)
